@@ -15,28 +15,35 @@
 """A CPython compatible SSLContext implementation wrapping PyOpenSSL's
 context.
 """
+from __future__ import annotations
 
+import asyncio
 import socket as _socket
 import ssl as _stdlibssl
 import sys as _sys
 import time as _time
 from errno import EINTR as _EINTR
 from ipaddress import ip_address as _ip_address
+from typing import TYPE_CHECKING, Any, Callable, Optional, TypeVar, Union
 
-from cryptography.x509 import load_der_x509_certificate as _load_der_x509_certificate
+import cryptography.x509 as x509
+import service_identity
 from OpenSSL import SSL as _SSL
 from OpenSSL import crypto as _crypto
-from service_identity import CertificateError as _SICertificateError
-from service_identity import VerificationError as _SIVerificationError
-from service_identity.pyopenssl import verify_hostname as _verify_hostname
-from service_identity.pyopenssl import verify_ip_address as _verify_ip_address
 
 from pymongo.errors import ConfigurationError as _ConfigurationError
-from pymongo.errors import _CertificateError
+from pymongo.errors import _CertificateError  # type:ignore[attr-defined]
 from pymongo.ocsp_cache import _OCSPCache
 from pymongo.ocsp_support import _load_trusted_ca_certs, _ocsp_callback
 from pymongo.socket_checker import SocketChecker as _SocketChecker
 from pymongo.socket_checker import _errno_from_exception
+from pymongo.write_concern import validate_boolean
+
+if TYPE_CHECKING:
+    from ssl import VerifyMode
+
+
+_T = TypeVar("_T")
 
 try:
     import certifi
@@ -67,25 +74,28 @@ _VERIFY_MAP = {
     _stdlibssl.CERT_REQUIRED: _SSL.VERIFY_PEER | _SSL.VERIFY_FAIL_IF_NO_PEER_CERT,
 }
 
-_REVERSE_VERIFY_MAP = dict((value, key) for key, value in _VERIFY_MAP.items())
+_REVERSE_VERIFY_MAP = {value: key for key, value in _VERIFY_MAP.items()}
 
 
 # For SNI support. According to RFC6066, section 3, IPv4 and IPv6 literals are
 # not permitted for SNI hostname.
-def _is_ip_address(address):
+def _is_ip_address(address: Any) -> bool:
     try:
         _ip_address(address)
         return True
-    except (ValueError, UnicodeError):  # noqa: B014
+    except (ValueError, UnicodeError):
         return False
 
 
-# According to the docs for Connection.send it can raise
+# According to the docs for socket.send it can raise
 # WantX509LookupError and should be retried.
 BLOCKING_IO_ERRORS = (_SSL.WantReadError, _SSL.WantWriteError, _SSL.WantX509LookupError)
+BLOCKING_IO_READ_ERROR = _SSL.WantReadError
+BLOCKING_IO_WRITE_ERROR = _SSL.WantWriteError
+BLOCKING_IO_LOOKUP_ERROR = _SSL.WantX509LookupError
 
 
-def _ragged_eof(exc):
+def _ragged_eof(exc: BaseException) -> bool:
     """Return True if the OpenSSL.SSL.SysCallError is a ragged EOF."""
     return exc.args == (-1, "Unexpected EOF")
 
@@ -94,12 +104,20 @@ def _ragged_eof(exc):
 # https://github.com/pyca/pyopenssl/issues/176
 # https://docs.python.org/3/library/ssl.html#notes-on-non-blocking-sockets
 class _sslConn(_SSL.Connection):
-    def __init__(self, ctx, sock, suppress_ragged_eofs):
+    def __init__(
+        self,
+        ctx: _SSL.Context,
+        sock: Optional[_socket.socket],
+        suppress_ragged_eofs: bool,
+        is_async: bool = False,
+    ):
         self.socket_checker = _SocketChecker()
         self.suppress_ragged_eofs = suppress_ragged_eofs
-        super(_sslConn, self).__init__(ctx, sock)
+        super().__init__(ctx, sock)
+        self._is_async = is_async
 
-    def _call(self, call, *args, **kwargs):
+    def _call(self, call: Callable[..., _T], *args: Any, **kwargs: Any) -> _T:
+        is_async = kwargs.pop("allow_async", True) and self._is_async
         timeout = self.gettimeout()
         if timeout:
             start = _time.monotonic()
@@ -107,6 +125,14 @@ class _sslConn(_SSL.Connection):
             try:
                 return call(*args, **kwargs)
             except BLOCKING_IO_ERRORS as exc:
+                # Do not retry if the connection is in non-blocking mode.
+                if is_async or timeout == 0:
+                    raise exc
+                # Check for closed socket.
+                if self.fileno() == -1:
+                    if timeout and _time.monotonic() - start > timeout:
+                        raise _socket.timeout("timed out") from None
+                    raise SSLError("Underlying socket has been closed") from None
                 if isinstance(exc, _SSL.WantReadError):
                     want_read = True
                     want_write = False
@@ -118,41 +144,42 @@ class _sslConn(_SSL.Connection):
                     want_write = True
                 self.socket_checker.select(self, want_read, want_write, timeout)
                 if timeout and _time.monotonic() - start > timeout:
-                    raise _socket.timeout("timed out")
+                    raise _socket.timeout("timed out") from None
                 continue
 
-    def do_handshake(self, *args, **kwargs):
-        return self._call(super(_sslConn, self).do_handshake, *args, **kwargs)
+    def do_handshake(self, *args: Any, **kwargs: Any) -> None:
+        kwargs["allow_async"] = False
+        return self._call(super().do_handshake, *args, **kwargs)
 
-    def recv(self, *args, **kwargs):
+    def recv(self, *args: Any, **kwargs: Any) -> bytes:
         try:
-            return self._call(super(_sslConn, self).recv, *args, **kwargs)
+            return self._call(super().recv, *args, **kwargs)
         except _SSL.SysCallError as exc:
             # Suppress ragged EOFs to match the stdlib.
             if self.suppress_ragged_eofs and _ragged_eof(exc):
                 return b""
             raise
 
-    def recv_into(self, *args, **kwargs):
+    def recv_into(self, *args: Any, **kwargs: Any) -> int:
         try:
-            return self._call(super(_sslConn, self).recv_into, *args, **kwargs)
+            return self._call(super().recv_into, *args, **kwargs)
         except _SSL.SysCallError as exc:
             # Suppress ragged EOFs to match the stdlib.
             if self.suppress_ragged_eofs and _ragged_eof(exc):
                 return 0
             raise
 
-    def sendall(self, buf, flags=0):
+    def sendall(self, buf: bytes, flags: int = 0) -> None:  # type: ignore[override]
         view = memoryview(buf)
         total_length = len(buf)
         total_sent = 0
         while total_sent < total_length:
             try:
-                sent = self._call(super(_sslConn, self).send, view[total_sent:], flags)
+                sent = self._call(super().send, view[total_sent:], flags)
             # XXX: It's not clear if this can actually happen. PyOpenSSL
             # doesn't appear to have any interrupt handling, nor any interrupt
             # errors for OpenSSL connections.
-            except (IOError, OSError) as exc:  # noqa: B014
+            except OSError as exc:
                 if _errno_from_exception(exc) == _EINTR:
                     continue
                 raise
@@ -163,23 +190,23 @@ class _sslConn(_SSL.Connection):
             total_sent += sent
 
 
-class _CallbackData(object):
+class _CallbackData:
     """Data class which is passed to the OCSP callback."""
 
-    def __init__(self):
-        self.trusted_ca_certs = None
-        self.check_ocsp_endpoint = None
+    def __init__(self) -> None:
+        self.trusted_ca_certs: Optional[list[x509.Certificate]] = None
+        self.check_ocsp_endpoint: Optional[bool] = None
         self.ocsp_response_cache = _OCSPCache()
 
 
-class SSLContext(object):
+class SSLContext:
     """A CPython compatible SSLContext implementation wrapping PyOpenSSL's
     context.
     """
 
     __slots__ = ("_protocol", "_ctx", "_callback_data", "_check_hostname")
 
-    def __init__(self, protocol):
+    def __init__(self, protocol: int):
         self._protocol = protocol
         self._ctx = _SSL.Context(self._protocol)
         self._callback_data = _CallbackData()
@@ -192,68 +219,80 @@ class SSLContext(object):
         self._ctx.set_ocsp_client_callback(callback=_ocsp_callback, data=self._callback_data)
 
     @property
-    def protocol(self):
+    def protocol(self) -> int:
         """The protocol version chosen when constructing the context.
         This attribute is read-only.
         """
         return self._protocol
 
-    def __get_verify_mode(self):
+    def __get_verify_mode(self) -> VerifyMode:
         """Whether to try to verify other peers' certificates and how to
         behave if verification fails. This attribute must be one of
         ssl.CERT_NONE, ssl.CERT_OPTIONAL or ssl.CERT_REQUIRED.
         """
         return _REVERSE_VERIFY_MAP[self._ctx.get_verify_mode()]
 
-    def __set_verify_mode(self, value):
+    def __set_verify_mode(self, value: VerifyMode) -> None:
         """Setter for verify_mode."""
 
-        def _cb(connobj, x509obj, errnum, errdepth, retcode):
+        def _cb(
+            _connobj: _SSL.Connection,
+            _x509obj: _crypto.X509,
+            _errnum: int,
+            _errdepth: int,
+            retcode: int,
+        ) -> bool:
             # It seems we don't need to do anything here. Twisted doesn't,
             # and OpenSSL's SSL_CTX_set_verify let's you pass NULL
             # for the callback option. It's weird that PyOpenSSL requires
             # this.
-            return retcode
+            # This is optional in pyopenssl >= 20 and can be removed once minimum
+            # supported version is bumped
+            # See: pyopenssl.org/en/latest/changelog.html#id47
+            return bool(retcode)
 
         self._ctx.set_verify(_VERIFY_MAP[value], _cb)
 
     verify_mode = property(__get_verify_mode, __set_verify_mode)
 
-    def __get_check_hostname(self):
+    def __get_check_hostname(self) -> bool:
         return self._check_hostname
 
-    def __set_check_hostname(self, value):
-        if not isinstance(value, bool):
-            raise TypeError("check_hostname must be True or False")
+    def __set_check_hostname(self, value: Any) -> None:
+        validate_boolean("check_hostname", value)
         self._check_hostname = value
 
     check_hostname = property(__get_check_hostname, __set_check_hostname)
 
-    def __get_check_ocsp_endpoint(self):
+    def __get_check_ocsp_endpoint(self) -> Optional[bool]:
         return self._callback_data.check_ocsp_endpoint
 
-    def __set_check_ocsp_endpoint(self, value):
-        if not isinstance(value, bool):
-            raise TypeError("check_ocsp must be True or False")
+    def __set_check_ocsp_endpoint(self, value: bool) -> None:
+        validate_boolean("check_ocsp", value)
         self._callback_data.check_ocsp_endpoint = value
 
     check_ocsp_endpoint = property(__get_check_ocsp_endpoint, __set_check_ocsp_endpoint)
 
-    def __get_options(self):
+    def __get_options(self) -> int:
         # Calling set_options adds the option to the existing bitmask and
         # returns the new bitmask.
         # https://www.pyopenssl.org/en/stable/api/ssl.html#OpenSSL.SSL.Context.set_options
         return self._ctx.set_options(0)
 
-    def __set_options(self, value):
-        # Explcitly convert to int, since newer CPython versions
+    def __set_options(self, value: int) -> None:
+        # Explicitly convert to int, since newer CPython versions
         # use enum.IntFlag for options. The values are the same
         # regardless of implementation.
         self._ctx.set_options(int(value))
 
     options = property(__get_options, __set_options)
 
-    def load_cert_chain(self, certfile, keyfile=None, password=None):
+    def load_cert_chain(
+        self,
+        certfile: Union[str, bytes],
+        keyfile: Union[str, bytes, None] = None,
+        password: Optional[str] = None,
+    ) -> None:
         """Load a private key and the corresponding certificate. The certfile
         string must be the path to a single file in PEM format containing the
         certificate as well as any number of CA certificates needed to
@@ -266,10 +305,11 @@ class SSLContext(object):
         # Password callback MUST be set first or it will be ignored.
         if password:
 
-            def _pwcb(max_length, prompt_twice, user_data):
+            def _pwcb(_max_length: int, _prompt_twice: bool, _user_data: Optional[bytes]) -> bytes:
                 # XXX:We could check the password length against what OpenSSL
                 # tells us is the max, but we can't raise an exception, so...
                 # warn?
+                assert password is not None
                 return password.encode("utf-8")
 
             self._ctx.set_passwd_cb(_pwcb)
@@ -277,7 +317,9 @@ class SSLContext(object):
         self._ctx.use_privatekey_file(keyfile or certfile)
         self._ctx.check_privatekey()
 
-    def load_verify_locations(self, cafile=None, capath=None):
+    def load_verify_locations(
+        self, cafile: Optional[str] = None, capath: Optional[str] = None
+    ) -> None:
         """Load a set of "certification authority"(CA) certificates used to
         validate other peers' certificates when `~verify_mode` is other than
         ssl.CERT_NONE.
@@ -285,9 +327,10 @@ class SSLContext(object):
         self._ctx.load_verify_locations(cafile, capath)
         # Manually load the CA certs when get_verified_chain is not available (pyopenssl<20).
         if not hasattr(_SSL.Connection, "get_verified_chain"):
+            assert cafile is not None
             self._callback_data.trusted_ca_certs = _load_trusted_ca_certs(cafile)
 
-    def _load_certifi(self):
+    def _load_certifi(self) -> None:
         """Attempt to load CA certs from certifi."""
         if _HAVE_CERTIFI:
             self.load_verify_locations(certifi.where())
@@ -299,18 +342,20 @@ class SSLContext(object):
                 "the tlsCAFile option"
             )
 
-    def _load_wincerts(self, store):
+    def _load_wincerts(self, store: str) -> None:
         """Attempt to load CA certs from Windows trust store."""
         cert_store = self._ctx.get_cert_store()
+        assert cert_store is not None
         oid = _stdlibssl.Purpose.SERVER_AUTH.oid
+
         for cert, encoding, trust in _stdlibssl.enum_certificates(store):  # type: ignore
             if encoding == "x509_asn":
                 if trust is True or oid in trust:
                     cert_store.add_cert(
-                        _crypto.X509.from_cryptography(_load_der_x509_certificate(cert))
+                        _crypto.X509.from_cryptography(x509.load_der_x509_certificate(cert))
                     )
 
-    def load_default_certs(self):
+    def load_default_certs(self) -> None:
         """A PyOpenSSL version of load_default_certs from CPython."""
         # PyOpenSSL is incapable of loading CA certs from Windows, and mostly
         # incapable on macOS.
@@ -326,23 +371,76 @@ class SSLContext(object):
             self._load_certifi()
         self._ctx.set_default_verify_paths()
 
-    def set_default_verify_paths(self):
+    def set_default_verify_paths(self) -> None:
         """Specify that the platform provided CA certificates are to be used
-        for verification purposes."""
+        for verification purposes.
+        """
         # Note: See PyOpenSSL's docs for limitations, which are similar
         # but not that same as CPython's.
         self._ctx.set_default_verify_paths()
 
+    async def a_wrap_socket(
+        self,
+        sock: _socket.socket,
+        server_side: bool = False,
+        do_handshake_on_connect: bool = True,
+        suppress_ragged_eofs: bool = True,
+        server_hostname: Optional[str] = None,
+        session: Optional[_SSL.Session] = None,
+    ) -> _sslConn:
+        """Wrap an existing Python socket connection and return a TLS socket
+        object.
+        """
+        ssl_conn = _sslConn(self._ctx, sock, suppress_ragged_eofs, True)
+        loop = asyncio.get_running_loop()
+        if session:
+            ssl_conn.set_session(session)
+        if server_side is True:
+            ssl_conn.set_accept_state()
+        else:
+            # SNI
+            if server_hostname and not _is_ip_address(server_hostname):
+                # XXX: Do this in a callback registered with
+                # SSLContext.set_info_callback? See Twisted for an example.
+                ssl_conn.set_tlsext_host_name(server_hostname.encode("idna"))
+            if self.verify_mode != _stdlibssl.CERT_NONE:
+                # Request a stapled OCSP response.
+                await loop.run_in_executor(None, ssl_conn.request_ocsp)
+            ssl_conn.set_connect_state()
+        # If this wasn't true the caller of wrap_socket would call
+        # do_handshake()
+        if do_handshake_on_connect:
+            # XXX: If we do hostname checking in a callback we can get rid
+            # of this call to do_handshake() since the handshake
+            # will happen automatically later.
+            await loop.run_in_executor(None, ssl_conn.do_handshake)
+            # XXX: Do this in a callback registered with
+            # SSLContext.set_info_callback? See Twisted for an example.
+            if self.check_hostname and server_hostname is not None:
+                from service_identity import pyopenssl
+
+                try:
+                    if _is_ip_address(server_hostname):
+                        pyopenssl.verify_ip_address(ssl_conn, server_hostname)
+                    else:
+                        pyopenssl.verify_hostname(ssl_conn, server_hostname)
+                except (  # type:ignore[misc]
+                    service_identity.SICertificateError,
+                    service_identity.SIVerificationError,
+                ) as exc:
+                    raise _CertificateError(str(exc)) from None
+        return ssl_conn
+
     def wrap_socket(
         self,
-        sock,
-        server_side=False,
-        do_handshake_on_connect=True,
-        suppress_ragged_eofs=True,
-        server_hostname=None,
-        session=None,
-    ):
-        """Wrap an existing Python socket sock and return a TLS socket
+        sock: _socket.socket,
+        server_side: bool = False,
+        do_handshake_on_connect: bool = True,
+        suppress_ragged_eofs: bool = True,
+        server_hostname: Optional[str] = None,
+        session: Optional[_SSL.Session] = None,
+    ) -> _sslConn:
+        """Wrap an existing Python socket connection and return a TLS socket
         object.
         """
         ssl_conn = _sslConn(self._ctx, sock, suppress_ragged_eofs)
@@ -370,11 +468,16 @@ class SSLContext(object):
             # XXX: Do this in a callback registered with
             # SSLContext.set_info_callback? See Twisted for an example.
             if self.check_hostname and server_hostname is not None:
+                from service_identity import pyopenssl
+
                 try:
                     if _is_ip_address(server_hostname):
-                        _verify_ip_address(ssl_conn, server_hostname)
+                        pyopenssl.verify_ip_address(ssl_conn, server_hostname)
                     else:
-                        _verify_hostname(ssl_conn, server_hostname)
-                except (_SICertificateError, _SIVerificationError) as exc:
-                    raise _CertificateError(str(exc))
+                        pyopenssl.verify_hostname(ssl_conn, server_hostname)
+                except (  # type:ignore[misc]
+                    service_identity.SICertificateError,
+                    service_identity.SIVerificationError,
+                ) as exc:
+                    raise _CertificateError(str(exc)) from None
         return ssl_conn
